@@ -41,51 +41,50 @@ import socket
 import time
 import email.utils
 import logging
-import rfc822
+import email
 import ssl
+import threading
+from http import client, server
 
-# Manage Python2/3 different modules
-# pylint: disable=F0401
-try:
-    from http import client, server
-except ImportError:
-    import httplib as client
-    import BaseHTTPServer as server
-# pylint: enable=F0401
 
-from . import acl, config, webdav, xmlutils, paths
+from . import acl, config, webdav, xmlutils, paths, gssapi
 
 log = logging.getLogger()
 ch = logging.StreamHandler()
 formatter = logging.Formatter("%(message)s")
 ch.setFormatter (formatter)
 log.addHandler(ch)
+negotiate = gssapi.Negotiate(log)
 
-VERSION = "1.5"
+VERSION = "2.0"
 
 def _check(request, function):
     """Check if user has sufficient rights for performing ``request``."""
     # ``_check`` decorator can access ``request`` protected functions
     # pylint: disable=W0212
+    owner = user = password = None
+    negotiate_success = False
 
-    authorization = request.headers.get("Authorization", None)
-    if authorization:
-        challenge = authorization.lstrip("Basic").strip().encode("ascii")
-        plain = request._decode(base64.b64decode(challenge))
-        user, password = plain.split(":")
-    else:
-        user = password = None
-
-    owner = None
     if request._collection:
         owner = request._collection.owner
 
+    authorization = request.headers.get("Authorization", None)
+    if authorization:
+        if authorization.startswith("Basic"):
+            challenge = authorization.lstrip("Basic").strip().encode("ascii")
+            plain = request._decode(base64.b64decode(challenge))
+            user, password = plain.split(":")
+        elif negotiate.enabled():
+            user, negotiate_success = negotiate.try_aaa(authorization, request, owner)
+
     # Also send UNAUTHORIZED if there's no collection. Otherwise one
     # could probe the server for (non-)existing collections.
-    if request.server.acl.has_right(owner, user, password):
+    if request.server.acl.has_right(owner, user, password) or negotiate_success:
         function(request, context={"user": user, "user-agent": request.headers.get("User-Agent", None)})
     else:
         request.send_calypso_response(client.UNAUTHORIZED, 0)
+        if negotiate.enabled():
+            request.send_header("WWW-Authenticate", "Negotiate")
         request.send_header(
             "WWW-Authenticate",
             'Basic realm="Calypso CalDAV/CardDAV server - password required"')
@@ -101,7 +100,7 @@ class HTTPServer(server.HTTPServer):
     # pylint: disable=W0231
     def __init__(self, address, handler):
         """Create server."""
-        server.HTTPServer.__init__(self, address, handler)
+        server.ThreadingHTTPServer.__init__(self, address, handler)
         self.acl = acl.load()
     # pylint: enable=W0231
 
@@ -110,18 +109,14 @@ class HTTPSServer(HTTPServer):
     """HTTPS server."""
     PROTOCOL = "https"
 
-    def __init__(self, address, handler):
-        """Create server by wrapping HTTP socket in an SSL socket."""
-        HTTPServer.__init__(self, address, handler)
-        self.socket = ssl.wrap_socket(
-            socket.socket(self.address_family, self.socket_type),
-            server_side=True,
-            certfile=os.path.expanduser(config.get("server", "certificate")),
-            keyfile=os.path.expanduser(config.get("server", "key")),
-            ssl_version=ssl.PROTOCOL_SSLv23)
-        self.server_bind()
-        self.server_activate()
+    def server_bind(self):
+        HTTPServer.server_bind(self)
 
+        context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
+        context.load_cert_chain(os.path.expanduser(config.get("server", "certificate")),
+                                keyfile=os.path.expanduser(config.get("server", "key")))
+
+        self.socket = context.wrap_socket(self.socket,server_side=True)
 
 class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
     """HTTP requests handler for WebDAV collections."""
@@ -129,7 +124,7 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
 
     # Decorator checking rights before performing request
     check_rights = lambda function: lambda request: _check(request, function)
-        
+
     # We do set Content-Length on all replies, so we can use HTTP/1.1
     # with multiple requests (as desired by the android CalDAV sync program
 
@@ -138,6 +133,21 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
     timeout = 90
 
     server_version = "Calypso/%s" % VERSION
+    queued_headers = {}
+
+    def queue_header(self, keyword, value):
+        self.queued_headers[keyword] = value
+
+    def end_headers(self):
+        """
+        Send out all queued headers and invoke or super classes
+        end_header.
+        """
+        if self.queued_headers:
+            for keyword, val in self.queued_headers.items():
+                self.send_header(keyword, val)
+            self.queued_headers = {}
+        return server.BaseHTTPRequestHandler.end_headers(self)
 
     def address_string(self):
         return str(self.client_address[0])
@@ -152,6 +162,9 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
         self.send_response(response)
         self.send_connection_header()
         self.send_header("Content-Length", length)
+        for header, value in config.items('headers'):
+            self.send_header(header, value)
+
 
     def handle_one_request(self):
         """Handle a single HTTP request.
@@ -210,7 +223,7 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
             log.error("Request timed out: %r", e)
             self.close_connection = 1
             return
-        except ssl.SSLError, x:
+        except ssl.SSLError as x:
             #an io error. Discard this connection
             log.error("SSL request error: %r", x.args[0])
             self.close_connection = 1
@@ -218,6 +231,7 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
 
 
     collections = {}
+    collections_lock = threading.Lock()
 
     @property
     def _collection(self):
@@ -225,9 +239,10 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
         path = paths.collection_from_path(self.path)
         if not path:
             return None
-        if not path in CollectionHTTPHandler.collections:
-            CollectionHTTPHandler.collections[path] = webdav.Collection(path)
-        return CollectionHTTPHandler.collections[path]
+        with CollectionHTTPHandler.collections_lock:
+            if not path in CollectionHTTPHandler.collections:
+                CollectionHTTPHandler.collections[path] = webdav.Collection(path)
+            return CollectionHTTPHandler.collections[path]
 
     def _decode(self, text):
         """Try to decode text according to various parameters."""
@@ -270,89 +285,96 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
 
         self._answer = ''
         answer_text = ''
+        code = client.OK
+        last_modified = ''
         try:
             item_name = paths.resource_from_path(self.path)
-            if item_name and self._collection:
-                # Get collection item
-                item = self._collection.get_item(item_name)
-                if item:
-                    if is_get:
-                        answer_text = item.text
-                    etag = item.etag
-                else:
-                    self.send_response(client.GONE)
-                    self.send_header("Content-Length", 0)
-                    self.end_headers()
-                    return
-            elif self._collection:
-                # Get whole collection
-                if is_get:
-                    answer_text = self._collection.text
-                etag = self._collection.etag
+            if self._collection:
+                with self._collection:
+                    if item_name:
+                        # Get collection item
+                        item = self._collection.get_item(item_name)
+                        if item:
+                            if is_get:
+                                answer_text = item.text
+                            etag = item.etag
+                        else:
+                            code = client.GONE
+                    else:
+                        # Get whole collection
+                        if is_get:
+                            answer_text = self._collection.text
+                        etag = self._collection.etag
+                    last_modified = self._collection.last_modified
+
+                if len(answer_text):
+                    try:
+                        self._answer = answer_text.encode(self._encoding,"xmlcharrefreplace")
+                    except UnicodeDecodeError:
+                        answer_text = answer_text.decode(errors="ignore")
+                        self._answer = answer_text.encode(self._encoding,"ignore")
             else:
-                self.send_calypso_response(client.NOT_FOUND, 0)
-                self.end_headers()
-                return
-                
-            if is_get:
-                try:
-                    self._answer = answer_text.encode(self._encoding,"xmlcharrefreplace")
-                except UnicodeDecodeError:
-                    answer_text = answer_text.decode(errors="ignore")
-                    self._answer = answer_text.encode(self._encoding,"ignore")
+                code = client.NOT_FOUND
 
-            self.send_calypso_response(client.OK, len(self._answer))
-            self.send_header("Content-Type", "text/calendar")
-            self.send_header("Last-Modified", email.utils.formatdate(time.mktime(self._collection.last_modified)))
-            self.send_header("ETag", etag)
-            self.end_headers()
-            if is_get:
-                self.wfile.write(self._answer)
         except Exception:
             log.exception("Failed HEAD for %s", self.path)
-            self.send_calypso_response(client.BAD_REQUEST, 0)
-            self.end_headers()
+            code = client.BAD_REQUEST
 
-    def if_match(self, item):
-        header = self.headers.get("If-Match", item.etag)
-        header = rfc822.unquote(header)
-        if header == item.etag:
-            return True
-        quoted = '"' + item.etag + '"'
-        if header == quoted:
-            return True
-        extraquoted = rfc822.quote(quoted)
-        if header == extraquoted:
-            return True
+        self.send_calypso_response(code, len(self._answer))
+        if code == client.OK:
+            self.send_header("Content-Type", "text/calendar")
+            self.send_header("Last-Modified", email.utils.formatdate(time.mktime(last_modified)))
+            self.send_header("ETag", etag)
+        self.end_headers()
+        if len(self._answer):
+            self.wfile.write(self._answer)
+
+    def if_match(self, items):
+        for item in items:
+            header = self.headers.get("If-Match", item.etag)
+            header = email.utils.unquote(header)
+            log.debug("header '%s' etag '%s'" % (header, item.etag))
+            if header == item.etag:
+                return True
+            quoted = '"' + item.etag + '"'
+            if header == quoted:
+                return True
+            extraquoted = email.utils.quote(quoted)
+            if header == extraquoted:
+                return True
         return False
 
     @check_rights
     def do_DELETE(self, context):
         """Manage DELETE request."""
+        self._answer = ''
+        code = client.NO_CONTENT
         try:
             item_name = paths.resource_from_path(self.path)
-            item = self._collection.get_item(item_name)
+            with self._collection:
+                items = self._collection.get_items(item_name)
+
+                if len(items) and self.if_match(items):
+                    # No ETag precondition or precondition verified, delete item
+                    self._answer = xmlutils.delete(self.path, self._collection, context=context)
+
+                elif not len(items):
+                    # Item does not exist
+                    code = client.NOT_FOUND
+                else:
+                    # No item or ETag precondition not verified, do not delete item
+                    code = client.PRECONDITION_FAILED
 
-            if item and self.if_match(item):
-                # No ETag precondition or precondition verified, delete item
-                self._answer = xmlutils.delete(self.path, self._collection, context=context)
-                
-                self.send_calypso_response(client.NO_CONTENT, len(self._answer))
-                self.send_header("Content-Type", "text/xml")
-                self.end_headers()
-                self.wfile.write(self._answer)
-            elif not item:
-                # Item does not exist
-                self.send_calypso_response(client.NOT_FOUND, 0)
-                self.end_headers()
-            else:
-                # No item or ETag precondition not verified, do not delete item
-                self.send_calypso_response(client.PRECONDITION_FAILED, 0)
-                self.end_headers()
         except Exception:
             log.exception("Failed DELETE for %s", self.path)
-            self.send_calypso_response(client.BAD_REQUEST, 0)
-            self.end_headers()
+            code = client.BAD_REQUEST
+
+        self.send_calypso_response(code, len(self._answer))
+        if len(self._answer):
+            self.send_header("Content-Type", "text/xml")
+        self.end_headers()
+        if len(self._answer):
+            self.wfile.write(self._answer)
 
     @check_rights
     def do_MKCALENDAR(self, context):
@@ -375,13 +397,23 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
         try:
             xml_request = self.xml_request
             log.debug("PROPFIND %s", xml_request)
-            self._answer = xmlutils.propfind(
-                self.path, xml_request, self._collection,
-                self.headers.get("depth", "infinity"),
-                context)
-            log.debug("PROPFIND ANSWER %s", self._answer)
+            depth = self.headers.get("depth", "infinity")
+            if depth != "infinity":
+                with self._collection:
+                    self._answer = xmlutils.propfind(
+                        self.path, xml_request, self._collection,
+                        depth, context)
+                status = client.MULTI_STATUS
+            else:
+                self._answer = xmlutils.propfind_deny()
+                status = client.FORBIDDEN
 
-            self.send_calypso_response(client.MULTI_STATUS, len(self._answer))
+            if len(self._answer) < 100:
+                log.debug("PROPFIND ANSWER %s", self._answer)
+            else:
+                log.debug("PROPFIND ANSWER len %d", len(self._answer))
+
+            self.send_calypso_response(status, len(self._answer))
             self.send_header("DAV", "1, calendar-access")
             self.send_header("Content-Type", "text/xml")
             self.end_headers()
@@ -401,39 +433,45 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
             log.exception("Failed SEARCH for %s", self.path)
             self.send_calypso_response(client.BAD_REQUEST, 0)
             self.end_headers()
-        
+
     @check_rights
     def do_PUT(self, context):
         """Manage PUT request."""
+
+        code = client.CREATED
+        etag = None
         try:
             item_name = paths.resource_from_path(self.path)
-            item = self._collection.get_item(item_name)
-            if not item or self.if_match(item):
+            with self._collection:
+                items = self._collection.get_items(item_name)
+                if not len(items) or self.if_match(items):
 
-                # PUT allowed in 3 cases
-                # Case 1: No item and no ETag precondition: Add new item
-                # Case 2: Item and ETag precondition verified: Modify item
-                # Case 3: Item and no Etag precondition: Force modifying item
-                webdav_request = self._decode(self.xml_request)
-                new_item = xmlutils.put(self.path, webdav_request, self._collection, context=context)
-                
-                log.debug("item_name %s new_name %s", item_name, new_item.name)
-                etag = new_item.etag
-                #log.debug("replacement etag %s", etag)
+                    # PUT allowed in 3 cases
+                    # Case 1: No item and no ETag precondition: Add new item
+                    # Case 2: Item and ETag precondition verified: Modify item
+                    # Case 3: Item and no Etag precondition: Force modifying item
+                    webdav_request = self._decode(self.xml_request)
+                    new_item = xmlutils.put(self.path, webdav_request, self._collection, context=context)
+
+                    log.debug("item_name %s new_name %s", item_name, new_item.name)
+                    etag = new_item.etag
+                    #log.debug("replacement etag %s", etag)
+
+                    code = client.CREATED
+                    etag = new_item.etag
+                else:
+                    #log.debug("Precondition failed")
+                    # PUT rejected in all other cases
+                    code = client.PRECONDITION_FAILED
 
-                self.send_calypso_response(client.CREATED, 0)
-                self.send_header("ETag", etag)
-                self.end_headers()
-            else:
-                #log.debug("Precondition failed")
-                # PUT rejected in all other cases
-                self.send_calypso_response(client.PRECONDITION_FAILED, 0)
-                self.end_headers()
         except Exception:
             log.exception('Failed PUT for %s', self.path)
-            self.send_calypso_response(client.BAD_REQUEST, 0)
-            self.end_headers()
+            code = client.BAD_REQUEST
 
+        self.send_calypso_response(code, 0)
+        if etag:
+            self.send_header('ETag', etag)
+        self.end_headers()
 
     @check_rights
     def do_REPORT(self, context):
@@ -441,8 +479,13 @@ class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
         try:
             xml_request = self.xml_request
             log.debug("REPORT %s %s", self.path, xml_request)
-            self._answer = xmlutils.report(self.path, xml_request, self._collection)
-            log.debug("REPORT ANSWER %s", self._answer)
+            with self._collection:
+                self._answer = xmlutils.report(self.path, xml_request, self._collection)
+
+            if len(self._answer) < 100:
+                log.debug("REPORT ANSWER %s" % self._answer)
+            else:
+                log.debug("REPORT ANSWER %d" % len(self._answer))
             self.send_calypso_response(client.MULTI_STATUS, len(self._answer))
             self.send_header("Content-Type", "text/xml")
             self.end_headers()
