@@ -30,15 +30,22 @@ import os
 import codecs
 import time
 import hashlib
-import glob
 import logging
 import tempfile
 import vobject
 import re
 import subprocess
-
+import vobject.base
+import iniparse
+import configparser
+import threading
+from configparser import RawConfigParser as ConfigParser
 from . import config, paths
 
+METADATA_FILENAME = ".calypso-collection"
+
+# pylint: enable=F0401
+
 #
 # Recursive search for 'name' within 'vobject'
 #
@@ -54,46 +61,42 @@ def find_vobject_value(vobject, name):
             return value
     return None
 
+
 class Item(object):
 
     """Internal item. Wraps a vObject"""
 
-    def __init__(self, text, name=None, path=None):
+    def __init__(self, text, name=None, path=None, parent_urlpath=None):
         """Initialize object from ``text`` and different ``kwargs``."""
 
         self.log = logging.getLogger(__name__)
-        try:
-            text = text.encode('utf8')
-        except UnicodeDecodeError:
-            text = text.decode('latin1').encode('utf-8')
+
+        # Preserve the original text as the contents of the object
+
+        self.text = text
 
         # Strip out control characters
 
-        text = re.sub(r"[\x01-\x09\x0b-\x1F\x7F]","",text)
+        text = re.sub(r"[\x01-\x09\x0b-\x1F\x7F]","", text)
 
-        try:
-            self.object = vobject.readOne(text)
-        except Exception:
-            self.log.exception("Parse error in %s %s", name, path)
-            raise
+        self.object = vobject.readOne(text)
 
-
-        if not self.object.contents.has_key('x-calypso-name'):
+        if 'x-calpyso-name' not in self.object.contents:
             if not name:
                 if self.object.name == 'VCARD' or self.object.name == 'VEVENT':
-                    if not self.object.contents.has_key('uid'):
-                        self.object.add('UID').value = hashlib.sha1(text).hexdigest()
+                    if 'uid' not in self.object.contents:
+                        self.object.add('UID').value = hashlib.sha1(text.encode('utf-8')).hexdigest()
                     name = self.object.uid.value
                 else:
                     for child in self.object.getChildren():
                         if child.name == 'VEVENT' or child.name == 'VCARD':
-                            if not child.contents.has_key('uid'):
-                                child.add('UID').value = hashlib.sha1(text).hexdigest()
+                            if 'uid' not in child.contents:
+                                child.add('UID').value = hashlib.sha1(text.encode('utf-8')).hexdigest()
                             name = child.uid.value
                             break
                     if not name:
-                        name = hashlib.sha1(text).hexdigest()
-                
+                        name = hashlib.sha1(text.encode('utf-8')).hexdigest()
+
             self.object.add("X-CALYPSO-NAME").value = name
         else:
             names = self.object.contents[u'x-calypso-name']
@@ -102,8 +105,12 @@ class Item(object):
 
         self.path = path
         self.name = self.object.x_calypso_name.value
+        if parent_urlpath.endswith("/"):
+            self.urlpath = parent_urlpath + self.name
+        else:
+            self.urlpath = "/".join([parent_urlpath, self.name])
         self.tag = self.object.name
-        self.etag = hashlib.sha1(text).hexdigest()
+        self.etag = hashlib.sha1(self.text.encode('utf-8')).hexdigest()
 
     @property
     def is_vcard(self):
@@ -112,6 +119,8 @@ class Item(object):
             return True
         if self.object.name == 'VEVENT':
             return False
+        if self.object.name == 'VTODO':
+            return False
         if self.object.name == 'VCALENDAR':
             return False
         for child in self.object.getChildren():
@@ -119,6 +128,8 @@ class Item(object):
                 return True
             if child.name == 'VEVENT':
                 return False
+            if child.name == 'VTODO':
+                return False
         return False
 
     @property
@@ -128,6 +139,8 @@ class Item(object):
             return False
         if self.object.name == 'VEVENT':
             return True
+        if self.object.name == 'VTODO':
+            return True
         if self.object.name == 'VCALENDAR':
             return True
         for child in self.object.getChildren():
@@ -135,6 +148,12 @@ class Item(object):
                 return False
             if child.name == 'VEVENT':
                 return True
+            if child.name == 'VTODO':
+                return True
+        return False
+
+    @property
+    def is_collection(self):
         return False
 
     @property
@@ -153,15 +172,6 @@ class Item(object):
             return '.ics'
         return '.dav'
 
-    @property
-    def text(self):
-        """Item text.
-
-        Text is the serialized form of the item.
-
-        """
-        return self.object.serialize().decode('utf-8')
-
     @property
     def length(self):
         return "%d" % len(self.text)
@@ -173,7 +183,16 @@ class Item(object):
             return value.utctimetuple()
         return time.gmtime()
 
-    def __unicode__(self):
+    def match_name(self, name):
+        if self.name == name:
+            return True
+        if self.name + self.file_extension == name:
+            return True
+        if self.name == name + self.file_extension:
+            return True
+        return False
+
+    def __str__(self):
         fn = self.object.getChildValue("fn")
         if fn:
             return fn
@@ -198,7 +217,6 @@ class Item(object):
 
         return self.name
 
-
 class Pathtime(object):
     """Path name and timestamps"""
 
@@ -217,6 +235,57 @@ class Pathtime(object):
         self.mtime = newmtime
         return False
 
+class Directory(Pathtime):
+
+    def __init__(self, path, name, parent_urlpath):
+        Pathtime.__init__(self, path)
+        self.name = name
+        self.etag = hashlib.sha1(path.encode('utf-8')).hexdigest()
+        if parent_urlpath.endswith("/"):
+            self.urlpath = parent_urlpath + self.name
+        else:
+            self.urlpath = "/".join([parent_urlpath, self.name])
+
+    @property
+    def is_vcard(self):
+        return False
+
+    @property
+    def is_vcal(self):
+        return False
+
+    @property
+    def is_collection(self):
+        return True
+
+    @property
+    def file_prefix(self):
+        return ''
+
+    @property
+    def file_extension(self):
+        return ''
+
+    @property
+    def text(self):
+        return self.path
+
+    @property
+    def length(self):
+        return "%d" % len(self.path)
+
+    @property
+    def last_modified(self):
+        return time.gmtime(self.curmtime)
+
+    @property
+    def is_calendar(self):
+        return True
+
+    @property
+    def is_addressbook(self):
+        return False
+
 class CalypsoError(Exception):
     def __init__(self, name, reason):
         self.name = name
@@ -230,24 +299,38 @@ class Collection(object):
 
     def get_description(self):
         try:
-            f = codecs.open(os.path.join(self.path, ".git/description"), encoding='utf-8')
-        except IOError:
+            return str(self.metadata.get('collection', 'description'))
+        except (configparser.NoSectionError, configparser.NoOptionError, ValueError):
+            pass
+
+        try:
+            with codecs.open(os.path.join(self.path, ".git/description"), encoding='utf-8') as file:
+                return file.read()
+        except:
             # .git/description is not present eg when the complete server is a single git repo
             return self.urlpath
-        return f.read()
 
     def read_file(self, path):
-        text = codecs.open(path,encoding='utf-8').read()
-        item = Item(text, None, path)
-        return item
+        with codecs.open(path,encoding='utf-8') as file:
+            text = file.read()
+            item = Item(text, None, path, self.urlpath)
+            return item
 
-    def insert_file(self, path):
-        try:
-            item = self.read_file(path)
-            self.my_items.append(item)
-        except Exception, ex:
-            self.log.exception("Insert %s failed", path)
-            return
+    def insert_file(self, path, name):
+        if os.path.isdir(path):
+            try:
+                item = Directory(path, name, self.urlpath)
+                self.my_dirs.append(item)
+            except Exception as ex:
+                self.log.exception("Insert %s failed", path)
+                return
+        else:
+            try:
+                item = self.read_file(path)
+                self.my_items.append(item)
+            except Exception as ex:
+                self.log.exception("Insert %s failed", path)
+                return
 
     def remove_file(self, path):
         old_items=[]
@@ -256,10 +339,32 @@ class Collection(object):
                 old_items.append(old_item)
         for old_item in old_items:
             self.my_items.remove(old_item)
-        
-    def scan_file(self, path):
+        old_dirs=[]
+        for old_dir in self.my_dirs:
+            if old_dir.path == path:
+                old_dirs.append(old_dir)
+        for old_dir in old_dirs:
+            self.my_dirs.remove(old_dir)
+
+    def scan_file(self, path, name):
         self.remove_file(path)
-        self.insert_file(path)
+        self.insert_file(path, name)
+
+    __metadatafile = property(lambda self: os.path.join(self.path, METADATA_FILENAME))
+
+    def scan_metadata(self, force):
+        try:
+            mtime = os.path.getmtime(self.__metadatafile)
+        except OSError:
+            mtime = 0
+            force = True
+
+        if not force and mtime == self.mtime and self.metadata is not None:
+            return
+
+        parser = ConfigParser()
+        parser.read(self.__metadatafile)
+        self.metadata = parser
 
     def scan_dir(self, force):
         try:
@@ -268,60 +373,87 @@ class Collection(object):
             mtime = 0
             force = True
 
+        self.scan_metadata(force)
+
         if not force and mtime == self.mtime:
             return
         self.log.debug("Scan %s", self.path)
         self.mtime = mtime
-        filenames = glob.glob(self.pattern)
+        filenames = os.listdir(self.path)
         newfiles = []
         for filename in filenames:
+            if filename == METADATA_FILENAME:
+                continue
+            if filename == '.git':
+                continue
+            filepath = os.path.join(self.path, filename)
             for file in self.files:
-                if filename == file.path:
+                if filepath == file.path:
                     newfiles.append(file)
                     if not file.is_up_to_date():
-                        self.log.debug("Changed %s", filename)
-                        self.scan_file(filename)
+                        self.log.debug("Changed %s", filepath)
+                        self.scan_file(filepath, filename)
                     break
             else:
-                if os.path.isdir(filename):
-                    self.log.debug("Ignoring directory %s in scan_dir", filename)
-                else:
-                    self.log.debug("New %s", filename)
-                    newfiles.append(Pathtime(filename))
-                    self.insert_file(filename)
+                self.log.debug("New %s", filepath)
+                newfiles.append(Pathtime(filepath))
+                self.insert_file(filepath, filename)
+
         for file in self.files:
-            if not file.path in filenames:
+            if os.path.basename(file.path) not in filenames:
                 self.log.debug("Removed %s", file.path)
                 self.remove_file(file.path)
+
         h = hashlib.sha1()
+        self._text = ''
         for item in self.my_items:
-            h.update(item.etag)
+            if getattr(item, 'etag', None):
+                h.update(item.etag.encode('utf-8'))
+            else:
+                h.update(item.ctag.encode('utf-8'))
+            self._text = self._text + item.text
         self._ctag = '%d-' % self.mtime + h.hexdigest()
         self.files = newfiles
-                
+
     def __init__(self, path):
         """Initialize the collection with ``cal`` and ``user`` parameters."""
-        
+
         self.log = logging.getLogger(__name__)
         self.encoding = "utf-8"
         self.urlpath = path
         self.owner = paths.url_to_owner(path)
         self.path = paths.url_to_file(path)
-        self.pattern = os.path.join(self.path, "*")
         self.files = []
         self.my_items = []
+        self.my_dirs = []
         self.mtime = 0
+        self._text = ''
         self._ctag = ''
-        self.etag = hashlib.sha1(self.path).hexdigest()
+        self.etag = hashlib.sha1(self.path.encode('utf-8')).hexdigest()
+        self.metadata = None
+        self.metadata_mtime = None
         self.scan_dir(False)
         self.tag = "Collection"
+        self.lock = threading.Lock()
 
     def __str__(self):
         return "Calendar-%s (at %s)" % (self.name, self.path)
 
     def __repr__(self):
         return "<Calendar %s>" % (self.name)
-        
+
+    def acquire(self):
+        self.lock.acquire()
+
+    def release(self):
+        self.lock.release()
+
+    def __enter__(self):
+        self.lock.__enter__()
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.lock.__exit__(exc_type, exc_val, exc_tb)
+
     def has_git(self):
         return True
 
@@ -351,7 +483,7 @@ class Collection(object):
         if self.has_git():
             subprocess.check_call(["git", "add", os.path.basename(path)], cwd=self.path)
             self.git_commit(context=context)
-    
+
     def git_rm(self, path, context):
         if self.has_git():
             subprocess.check_call(["git", "rm", os.path.basename(path)], cwd=self.path)
@@ -364,14 +496,14 @@ class Collection(object):
             # Touch directory so that another running instance will update
             try:
                 os.utime(self.path, None)
-            except Exception, ex:
+            except Exception as ex:
                 self.log.exception("Failed to set directory mtime")
-            
+
     def write_file(self, item):
         fd, path = tempfile.mkstemp(item.file_extension, item.file_prefix, dir=self.path)
         self.log.debug('Trying to write to %s', path)
         file = os.fdopen(fd, 'w')
-        file.write(item.text.encode('utf-8'))
+        file.write(item.text)
         file.close()
         self.log.debug('Wrote %s to %s', file, path)
         return path
@@ -382,7 +514,7 @@ class Collection(object):
         if not os.path.exists(os.path.dirname(self.path)):
             try:
                 os.makedirs(os.path.dirname(self.path))
-            except OSError, ose:
+            except OSError as ose:
                 self.log.exception("Failed to make collection directory %s: %s", self.path, ose)
                 raise
 
@@ -392,10 +524,10 @@ class Collection(object):
             path = self.write_file(item)
             self.git_add(path, context=context)
             self.scan_dir(True)
-        except OSError, ex:
+        except OSError as ex:
             self.log.exception("Error writing file")
             raise
-        except Exception, ex:
+        except Exception as ex:
             self.log.exception("Caught Exception")
             self.log.debug("Failed to create %s: %s", self.path,  ex)
             raise
@@ -409,7 +541,7 @@ class Collection(object):
             os.unlink(item.path)
             self.git_rm(item.path, context=context)
             self.scan_dir(True)
-        except Exception, ex:
+        except Exception as ex:
             self.log.exception("Failed to remove %s", item.path)
             raise
 
@@ -421,17 +553,17 @@ class Collection(object):
         try:
             new_path = self.write_file(item)
             os.rename(new_path, item.path)
-            self.scan_file(item.path)
+            self.scan_file(item.path, item.name)
             self.git_change(item.path, context=context)
             self.scan_dir(True)
-        except Exception, ex:
+        except Exception as ex:
             self.log.exception("Failed to rewrite %s", item.path)
             raise
-        
+
     def get_item(self, name):
         """Get collection item called ``name``."""
         for item in self.my_items:
-            if item.name == name:
+            if item.match_name(name):
                 return item
         return None
 
@@ -439,7 +571,7 @@ class Collection(object):
         """Get collection items called ``name``."""
         items=[]
         for item in self.my_items:
-            if item.name == name:
+            if item.match_name(name):
                 items.append(item)
         return items
 
@@ -452,12 +584,13 @@ class Collection(object):
 
         self.log.debug('append name %s', name)
         try:
-            new_item = Item(text, name, None)
-        except Exception, e:
+            new_item = Item(text, name, None, self.urlpath)
+        except Exception as e:
             self.log.exception("Cannot create new item")
             raise
-        if new_item.name in (item.name for item in self.my_items):
-            self.log.debug("Item %s already present %s" , new_item.name, self.get_item(new_item.name).path)
+        old_item = self.get_item(new_item.name)
+        if old_item:
+            self.log.debug("Item %s already present %s" , new_item.name, old_item.path)
             raise CalypsoError(new_item.name, "Item already present")
         self.log.debug("New item %s", new_item.name)
         self.create_file(new_item, context=context)
@@ -467,9 +600,9 @@ class Collection(object):
         """Remove object named ``name`` from collection."""
         self.log.debug("Remove object %s", name)
         for old_item in self.my_items:
-            if old_item.name == name:
+            if old_item.match_name(name):
                 self.destroy_file(old_item, context=context)
-                
+
     def replace(self, name, text, context):
         """Replace content by ``text`` in objet named ``name`` in collection."""
 
@@ -479,7 +612,7 @@ class Collection(object):
             path = old_item.path
 
         try:
-            new_item = Item(text, name, path)
+            new_item = Item(text, name, path, self.urlpath)
         except Exception:
             self.log.exception("Failed to replace %s", name)
             raise
@@ -509,30 +642,35 @@ class Collection(object):
         """
 
         try:
-            new_object = vobject.readComponents(codecs.open(path,encoding='utf-8').read())
-            for new_ics in new_object:
-                if new_ics.name == 'VCALENDAR':
+            with codecs.open(path,encoding='utf-8') as file:
+                new_object = vobject.readComponents(file.read())
+                for new_ics in new_object:
+                    if new_ics.name == 'VCALENDAR':
 
-                    events = new_ics.vevent_list
-                    for ve in events:
-                        # Check for events with both dtstart and duration entries and
-                        # delete the duration one
-                        if ve.contents.has_key('dtstart') and ve.contents.has_key('duration'):
-                            del ve.contents['duration']
-                        new_ics.vevent_list = [ve]
-                        new_item = Item(new_ics.serialize().decode('utf-8'), None, path)
+                        events = new_ics.vevent_list
+                        for ve in events:
+                            # Check for events with both dtstart and duration entries and
+                            # delete the duration one
+                            if 'dtstart' in ve.contents and 'duration' in ve.contents:
+                                del ve.contents['duration']
+                            new_ics.vevent_list = [ve]
+                            new_item = Item(new_ics.serialize(), None, path, self.urlpath)
+                            self.import_item(new_item, path)
+                    else:
+                        new_item = Item(new_ics.serialize(), None, path, self.urlpath)
                         self.import_item(new_item, path)
-                else:
-                    new_item = Item(new_ics.serialize().decode('utf-8'), None, path)
-                    self.import_item(new_item, path)
             return True
-        except Exception, ex:
+        except Exception as ex:
             self.log.exception("Failed to import: %s", path)
             return False
-        
+
     def write(self, headers=None, items=None):
         return True
 
+    @property
+    def is_collection(self):
+        return True
+
     @property
     def ctag(self):
         self.scan_dir(False)
@@ -548,10 +686,15 @@ class Collection(object):
     def text(self):
         """Collection as plain text."""
         self.scan_dir(False)
-        _text = ""
-        for item in self.my_items:
-            _text = _text + item.text
-        return _text
+        return self._text
+
+    @property
+    def color(self):
+        """Color."""
+        try:
+            return "#%s" % self.metadata.get('collection', 'color')
+        except (configparser.NoSectionError, configparser.NoOptionError, ValueError):
+            return None
 
     @property
     def headers(self):
@@ -564,6 +707,12 @@ class Collection(object):
         self.scan_dir(False)
         return self.my_items
 
+    @property
+    def dirs(self):
+        """Get list of all dirs in collection."""
+        self.scan_dir(False)
+        return self.my_dirs
+
     @property
     def last_modified(self):
         """Get the last time the collection has been modified.
@@ -577,3 +726,17 @@ class Collection(object):
     @property
     def length(self):
         return "%d" % len(self.text)
+
+    @property
+    def is_addressbook(self):
+        try:
+            return self.metadata.getboolean('collection', 'is-addressbook')
+        except (configparser.NoSectionError, configparser.NoOptionError, ValueError):
+            return True
+
+    @property
+    def is_calendar(self):
+        try:
+            return self.metadata.getboolean('collection', 'is-calendar')
+        except (configparser.NoSectionError, configparser.NoOptionError, ValueError):
+            return True
