@@ -46,6 +46,7 @@ NAMESPACES = {
     "C": "urn:ietf:params:xml:ns:caldav",
     "A": "urn:ietf:params:xml:ns:carddav",
     "D": "DAV:",
+    "E": "http://apple.com/ns/ical/",
     "CS": "http://calendarserver.org/ns/"}
 
 log = logging.getLogger(__name__)
@@ -117,14 +118,14 @@ def propfind(path, xml_request, collection, depth, context):
                  _tag("D", "getcontentlength"),
                  _tag("D", "getlastmodified")]
 
-    
+
     # Writing answer
     multistatus = ET.Element(_tag("D", "multistatus"))
 
     if collection:
         if item_name:
             item = collection.get_item(item_name)
-            print "item_name %s item %s" % (item_name, item)
+            log.debug("item_name %s item %s" % (item_name, item))
             if item:
                 items = [item]
             else:
@@ -133,20 +134,19 @@ def propfind(path, xml_request, collection, depth, context):
             if depth == "0":
                 items = [collection]
             else:
-                # depth is 1, infinity or not specified
-                # we limit ourselves to depth == 1
-                items = [collection] + collection.items
+                # We limit ourselves to depth == 1
+                items = [collection] + collection.items + collection.dirs
     else:
         items = []
 
     for item in items:
-        is_collection = isinstance(item, webdav.Collection)
+        is_collection = item.is_collection
 
         response = ET.Element(_tag("D", "response"))
         multistatus.append(response)
 
         href = ET.Element(_tag("D", "href"))
-        href.text = collection_name if is_collection else "/".join([collection_name, item.name])
+        href.text = item.urlpath
         response.append(href)
 
         propstat = ET.Element(_tag("D", "propstat"))
@@ -158,10 +158,12 @@ def propfind(path, xml_request, collection, depth, context):
         for tag in props:
             element = ET.Element(tag)
             if tag == _tag("D", "resourcetype") and is_collection:
-                tag = ET.Element(_tag("C", "calendar"))
-                element.append(tag)
-                tag = ET.Element(_tag("A", "addressbook"))
-                element.append(tag)
+                if item.is_calendar:
+                    tag = ET.Element(_tag("C", "calendar"))
+                    element.append(tag)
+                if item.is_addressbook:
+                    tag = ET.Element(_tag("A", "addressbook"))
+                    element.append(tag)
                 tag = ET.Element(_tag("D", "collection"))
                 element.append(tag)
             elif tag == _tag("D", "owner"):
@@ -176,7 +178,9 @@ def propfind(path, xml_request, collection, depth, context):
             elif tag == _tag("D", "getetag"):
                 element.text = item.etag
             elif tag == _tag("D", "displayname") and is_collection:
-                element.text = collection.name
+                element.text = item.name
+            elif tag == _tag("E", "calendar-color") and is_collection:
+                element.text = item.color
             elif tag == _tag("D", "principal-URL"):
                 # TODO: use a real principal URL, read rfc3744-4.2 for info
                 tag = ET.Element(_tag("D", "href"))
@@ -218,7 +222,7 @@ def propfind(path, xml_request, collection, depth, context):
                 element.append(tag)
             elif tag in (_tag("A", "addressbook-description"),
                          _tag("C", "calendar-description")) and is_collection:
-                element.text = collection.get_description()
+                element.text = item.get_description()
             prop.append(element)
 
         status = ET.Element(_tag("D", "status"))
@@ -228,11 +232,23 @@ def propfind(path, xml_request, collection, depth, context):
     return ET.tostring(multistatus, config.get("encoding", "request"))
 
 
+def propfind_deny():
+    """Answer an infinity PROPFIND requests.
+
+    Read rfc4918-9.1.1 for info.
+    """
+    error = ET.Element(_tag("D", "error"))
+    prec_code = ET.Element(_tag("D", "propfind-finite-depth"))
+    error.append(prec_code)
+    return ET.tostring(error, config.get("encoding", "request"))
+
+
 def put(path, webdav_request, collection, context):
     """Read PUT requests."""
     name = paths.resource_from_path(path)
     log.debug('xmlutils put path %s name %s', path, name)
-    if name in (item.name for item in collection.items):
+    old_item = collection.get_item(name)
+    if old_item:
         # PUT is modifying an existing item
         log.debug('Replacing item named %s', name)
         return collection.replace(name, webdav_request, context=context)
@@ -249,7 +265,7 @@ def match_filter_element(vobject, fe):
             if comp == vobject.name:
                 hassub = False
                 submatch = False
-                for fc in fe.getchildren():
+                for fc in fe:
                     if match_filter_element(vobject, fc):
                         submatch = True
                         break
@@ -318,7 +334,7 @@ def match_filter(item, filter):
         return True
     if filter.tag != _tag("C", "filter"):
         return True
-    for fe in filter.getchildren():
+    for fe in filter:
         if match_filter_element(item.object, fe):
             return True
     return False
@@ -357,17 +373,19 @@ def report(path, xml_request, collection):
         if name:
             # Reference is an item
             path = paths.collection_from_path(hreference) + "/"
-            items = (item for item in collection.items if item.name == name)
+            items = collection.get_items(name)
         else:
             # Reference is a collection
             path = hreference
             items = collection.items
 
-        
+
         for item in items:
             if not match_filter(item, filter_element):
                 continue
 
+            log.debug("match %s" % item)
+
             response = ET.Element(_tag("D", "response"))
             multistatus.append(response)
 
@@ -387,6 +405,8 @@ def report(path, xml_request, collection):
                     element.text = item.etag
                 elif tag == _tag("C", "calendar-data"):
                     element.text = item.text
+                elif tag == _tag("A", "address-data"):
+                    element.text = item.text
                 prop.append(element)
 
             status = ET.Element(_tag("D", "status"))
@@ -394,5 +414,5 @@ def report(path, xml_request, collection):
             propstat.append(status)
 
     reply = ET.tostring(multistatus, config.get("encoding", "request"))
-        
+
     return reply
