@@ -1,35 +1,41 @@
 # vim: set fileencoding=utf-8 :
-"""Test L{gbp.command_wrappers.Command}'s tarball unpack"""
 
 import subprocess
 import tempfile
 import shutil
 import unittest
 
-import calypso.config
 from calypso.webdav import Collection
 from calypso import paths
 
+from .testutils import CalypsoTestCase
 
-class TestCollection(unittest.TestCase):
+
+class TestCollection(CalypsoTestCase):
     test_vcard = "tests/data/import.vcard"
     test_resource_with_slash = "tests/data/from-tripsync.ics"
 
-    def setUp(self):
-        self.tmpdir = tempfile.mkdtemp()
-        calypso.config.set('storage', 'folder', self.tmpdir)
-        subprocess.call(["git", "init", self.tmpdir]),
-
-    def tearDown(self):
-        if self.tmpdir:
-            shutil.rmtree(self.tmpdir)
-
     def test_import_file(self):
         collection = Collection("")
         self.assertTrue(collection.import_file(self.test_vcard))
         self.assertEqual(len(collection.items), 2)
         org = u'Universitetet i Tromsø'
-        self.assertTrue(org == collection.items[0].object.org.value[0])
+        self.assertEqual(org, collection.items[0].object.org.value[0])
+
+    def test_remove_existent_item(self):
+        collection = Collection("")
+        self.assertTrue(collection.import_file(self.test_vcard))
+        self.assertEqual(len(collection.items), 2)
+        name = collection.items[0].name
+        collection.remove(name, {})
+        self.assertEqual(len(collection.items), 1)
+
+    def test_remove_nonexistent_item(self):
+        collection = Collection("")
+        self.assertTrue(collection.import_file(self.test_vcard))
+        self.assertEqual(len(collection.items), 2)
+        collection.remove("doesnotexist", {})
+        self.assertEqual(len(collection.items), 2)
 
     def test_uid_with_slash(self):
         collection = Collection("/")
@@ -38,5 +44,5 @@ class TestCollection(unittest.TestCase):
         veventuid = collection.items[0].object.vevent.uid.value
         r = paths.resource_from_path(veventuid)
         c = paths.collection_from_path(veventuid)
-        self.assertTrue(r == veventuid)
-        self.assertTrue("/" == c)
+        self.assertEqual(r, veventuid)
+        self.assertEqual("/", c)
